/*
 * Copyright 2002 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/queue.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <err.h>
#include <assert.h>

#include "dag.h"

#define DIFF(a,b) do { \
	if (a < b) return (-1); \
	if (a > b) return (1); \
} while (0);

#define SIGN(x)	(x < 0 ? -1 : 1)

int
nodecompare(struct dag_node *a, struct dag_node *b)
{
	DIFF(a->p_x, b->p_x);
	DIFF(a->p_y, b->p_y);
	return (0);

}

int
nodeidcompare(struct dag_node *a, struct dag_node *b)
{
	DIFF(a->id, b->id);
	return (0);

}

SPLAY_HEAD(nodetree, dag_node);
SPLAY_HEAD(nodeidtree, dag_node);

SPLAY_PROTOTYPE(nodetree, dag_node, node, nodecompare);
SPLAY_GENERATE(nodetree, dag_node, node, nodecompare);

SPLAY_PROTOTYPE(nodeidtree, dag_node, id_node, nodeidcompare);
SPLAY_GENERATE(nodeidtree, dag_node, id_node, nodeidcompare);

int
edgecompare(struct dag_edge *a, struct dag_edge *b)
{
	DIFF(a->destination, b->destination);
	return (0);
}

SPLAY_GENERATE(edgetree, dag_edge, node, edgecompare);

static TAILQ_HEAD(daglist, dag_node) daglist;
static struct nodetree noderoot;
static struct nodeidtree nodeidroot;
static int node_ids;
static int maxx, maxy;
static double maxforce;
static double stepsize = 0.08;

/* Tweakable parameters */
static double repulse_const = 150.0;
static double attract_const = 0.03;

static double min_repulsedist = 0.0001;
static double max_attractdist = 100;
static double repulse_center = 0.5;
static double attract_center = -0.7;
static double radius;
static int need_rehash = 0;

#define MAX_BUCKETS	128
static int xdivide;
static int ydivide;

struct daglist buckets[MAX_BUCKETS][MAX_BUCKETS];

void
dag_hash_init(void)
{
	 xdivide = (maxx + MAX_BUCKETS - 2)/(MAX_BUCKETS - 1);
	 ydivide = (maxy + MAX_BUCKETS - 2)/(MAX_BUCKETS - 1);
}

void
dag_init(void)
{
	int i, j;

	/* Init 2d buckets */
	for (i = 0; i < MAX_BUCKETS; i++) {
		for (j = 0; j < MAX_BUCKETS; j++)
			TAILQ_INIT(&buckets[i][j]);
	}
	
	TAILQ_INIT(&daglist);
	SPLAY_INIT(&noderoot);
	SPLAY_INIT(&nodeidroot);

	dag_hash_init();
}

double
dag_set_stepsize(double stepsz)
{
	double oldstep = stepsize;

	if (stepsz > 0.0001)
		stepsize = stepsz;

	return (oldstep);
}

void
dag_set_params(double repdist, double attrdist,
    double repcenter, double attrcenter)
{
	min_repulsedist = repdist;
	max_attractdist = attrdist;
	repulse_center = repcenter;
	attract_center = -attrcenter;
}

void
dag_get_size(int *pmaxx, int *pmaxy)
{
	*pmaxx = maxx;
	*pmaxy = maxy;
}

void
dag_get_params(double *prepdist, double *pattrdist,
    double *prepcenter, double *pattrcenter)
{
	*prepdist = min_repulsedist;
	*pattrdist = max_attractdist;
	*prepcenter = repulse_center;
	*pattrcenter = -attract_center;
}

void
dag_set_physics(double repulse, double attract)
{
	repulse_const = repulse;
	attract_const = attract;
}

void
dag_get_physics(double *prepulse, double *pattract)
{
	*prepulse = repulse_const;
	*pattract = attract_const;
}


struct dag_node *
dag_new_node(int x, int y)
{
	struct dag_node *node, tmp;

	tmp.p_x = x;
	tmp.p_y = y;

	if (SPLAY_FIND(nodetree, &noderoot, &tmp) != NULL)
		return (NULL);

	if (x > maxx)
		maxx = x;
	if (y > maxy)
		maxy = y;

	if ((node = calloc(1, sizeof(struct dag_node))) == NULL)
		err(1, "calloc");

	SPLAY_INIT(&node->edgeroot);

	node->p_x = x;
	node->p_y = y;
	node->id = node_ids++;

	need_rehash = 1;

	return (node);
}

struct dag_node *
dag_node_findid(int id)
{
	struct dag_node tmp;

	tmp.id = id;

	return (SPLAY_FIND(nodeidtree, &nodeidroot, &tmp));
}

void
dag_insert_node(struct dag_node *node)
{
	TAILQ_INSERT_TAIL(&daglist, node, next);
	SPLAY_INSERT(nodetree, &noderoot, node);

	SPLAY_INSERT(nodeidtree, &nodeidroot, node);
}

void
dag_remove_node(struct dag_node *node)
{
	struct dag_edge *edge, *next;

	for (edge = SPLAY_ROOT(&node->edgeroot); edge; edge = next) {
		next = SPLAY_NEXT(edgetree, &node->edgeroot, edge);

		dag_deconnect(node, edge->destination);
	}

	TAILQ_REMOVE(&daglist, node, next);
	SPLAY_REMOVE(nodetree, &noderoot, node);
	SPLAY_REMOVE(nodeidtree, &nodeidroot, node);

	need_rehash = 1;
}

void
dag_free_node(struct dag_node *node)
{
	free(node);
}

int
dag_walk_all(int (*cb)(struct dag_node *, void *), void *arg)
{
	struct dag_node *node;

	SPLAY_FOREACH(node, nodeidtree, &nodeidroot) {
		if (cb(node, arg) == -1)
			return (-1);
	}

	return (0);
}

int
dag_in_rectangle(struct dag_point *point, int x1, int y1, int x2, int y2)
{
	if (x1 > point->x || x2 < point->x)
		return (0);
	if (y1 > point->y || y2 < point->y)
		return (0);

	return (1);
}

#define CLIP(p, m) do { \
	if ((p) < 0) (p) = 0; \
	if ((p) > m) (p) = m; \
} while (0)

void
dag_node_hash(struct dag_point *point, int *px, int *py)
{
	*px = point->x / xdivide;
	*py = point->y / ydivide;
}

int
dag_walk_region(int (*cb)(struct dag_node *, void *),
    int x1, int y1, int x2, int y2, void *arg)
{
	struct dag_node *node;
	struct dag_point point;

	int xstart, xend;
	int ystart, yend;

	int x, y;

	CLIP(x1, maxx);
	CLIP(x2, maxx);
	CLIP(y1, maxy);
	CLIP(y2, maxy);

	point.x = x1;
	point.y = y1;
	dag_node_hash(&point, &xstart, &ystart);
	point.x = x2;
	point.y = y2;
	dag_node_hash(&point, &xend, &yend);

	for (x = xstart; x <= xend; x++) {
		for (y = ystart; y <= yend; y++) {
			struct daglist *list;

			list = &buckets[x][y];
			
			TAILQ_FOREACH(node, list, bucket_next) {
				if (!dag_in_rectangle(&node->point,
					x1, y1, x2, y2))
					continue;

				if (cb(node, arg) == -1)
					return (-1);
			}
		}
	}

	return (0);
}

struct dag_node *
dag_random_node(void)
{
	struct dag_node *node, *take = NULL;
	int count = 1;

	TAILQ_FOREACH(node, &daglist, next) {
		if (arc4random() % count == 0)
			take = node;

		count += 2;
	}

	return (take);
}

void
dag_hash(void)
{
	struct dag_node *node;
	int offx, offy;
	int i, j;

	if (!need_rehash)
		return;

	need_rehash = 0;

	dag_hash_init();

	/* Init 2d buckets */
	for (i = 0; i < MAX_BUCKETS; i++) {
		for (j = 0; j < MAX_BUCKETS; j++)
			TAILQ_INIT(&buckets[i][j]);
	}

	TAILQ_FOREACH(node, &daglist, next) {
		dag_node_hash(&node->point, &offx, &offy);

		TAILQ_INSERT_TAIL(&buckets[offx][offy], node, bucket_next);
	}
}


struct dag_edge *
dag_new_edge(struct dag_node *node)
{
	struct dag_edge *edge;

	edge = calloc(1, sizeof(struct dag_edge));
	if (edge == NULL)
		err(1, "calloc");

	edge->destination = node;

	return (edge);
}

void
dag_free_edge(struct dag_node *node, struct dag_edge *edge)
{
	SPLAY_REMOVE(edgetree, &node->edgeroot, edge);

	free(edge);
}

struct dag_edge *
dag_find_edge(struct dag_node *from, struct dag_node *which)
{
	struct dag_edge tmp;

	tmp.destination = which;

	return (SPLAY_FIND(edgetree, &from->edgeroot, &tmp));
}

void
dag_connect(struct dag_node *from, struct dag_node *dest, int undirected)
{
	struct dag_edge *one, *two;

	if (dag_isconnected(from, dest))
		return;

	one = dag_new_edge(dest);
	SPLAY_INSERT(edgetree, &from->edgeroot, one);
	assert(dag_find_edge(from, dest) != NULL);

	from->size++;

	if (undirected) {
		two = dag_new_edge(from);
		SPLAY_INSERT(edgetree, &dest->edgeroot, two);
		assert(dag_find_edge(dest, from) != NULL);

		dest->size++;

		one->undirected = two;
		two->undirected = one;
	}
}

void
dag_deconnect(struct dag_node *from, struct dag_node *dest)
{
	struct dag_edge *one, *two;

	if (!dag_isconnected(from, dest))
		return;

	one = dag_find_edge(from, dest);
	if (one == NULL)
		errx(1, "%s: dag_get_edge", __func__);

	two = one->undirected;

	dag_free_edge(from, one);
	if (two != NULL)
		dag_free_edge(dest, two);
}

int
dag_isconnected(struct dag_node *from, struct dag_node *dest)
{
	return (dag_find_edge(from, dest) != NULL);
}

#define SQ(x)	((x)*(x))

double
dag_distance(struct dag_node *one, struct dag_node *two)
{
	double dist;

	dist = SQ(one->p_x - two->p_x);
	dist += SQ(one->p_y - two->p_y);

	return (sqrt(dist));
}

struct dag_point *
dag_repulse(struct dag_node *one, struct dag_node *two, double min)
{
	static struct dag_point vector;
	double dist = dag_distance(one, two);
	double fact;

	if (dist < min)
		dist = min;

	vector.x = (one->p_x - two->p_x)/dist;
	vector.y = (one->p_y - two->p_y)/dist;

	fact = repulse_const * one->size * two->size / (dist * dist);

	vector.x *= fact;
	vector.y *= fact;

	return (&vector);
}

int
dag_compute_repulse(struct dag_node *from, struct dag_node *dest, double min)
{
	struct dag_point *point;

	if (dest == from)
		return (0);

	point = dag_repulse(from, dest, min);

	from->force.x += point->x;
	from->force.y += point->y;

	dest->force.x -= point->x;
	dest->force.y -= point->y;

	return (0);

}

int
dag_compute_repulse_cb(struct dag_node *node, void *arg)
{
	struct dag_node *from = arg;

	if (from < node)
		return (0);

	return (dag_compute_repulse(from, node, min_repulsedist));
}

int
dag_compute_attraction(struct dag_node *one, struct dag_node *two, double max)
{
	struct dag_point vector;
	double dist = dag_distance(one, two);
	double fact, size, size_one, size_two;

	if (dist < 0.0001)
		dist = 0.0001;

	vector.x = (two->p_x - one->p_x)/dist;
	vector.y = (two->p_y - one->p_y)/dist;

	if (dist > max)
		dist = max;

	size_one = 2*abs(one->size - two->size) + 1;
	size_two = one->size > two->size ? two->size : one->size;

	size = (size_one + size_two)/2;

	fact = attract_const * size * dist;

	vector.x *= fact;
	vector.y *= fact;

	one->force.x += vector.x;
	one->force.y += vector.y;

	/* Optimize */
	two->force.x -= vector.x;
	two->force.y -= vector.y;

	return (0);
}

#define C_TOP		1
#define C_BOTTOM	2
#define C_RIGHT		4
#define C_LEFT		8

int
ClipCode(struct dag_point *me, struct dag_point *max)
{
	int c = 0;

	if (me->y > max->y)
		c |= C_TOP;
	else if (me->y < 0)
		c |= C_BOTTOM;
	if (me->x > max->x)
		c |= C_RIGHT;
	else if (me->x < 0)
		c |= C_LEFT;

	return (c);
}

struct dag_point *
dag_cohensutherland_clip (struct dag_point *start, struct dag_point *end,
    struct dag_point *max)
{
	static struct dag_point result;

	int C;
	double x, y;

	C = ClipCode(end, max);

	/* Nothing to be clipped */
	if (C == 0)
		return (NULL);

	do {
		if (C & C_TOP) {
			x = start->x + (end->x - start->x) *
			    (max->y - start->y) / (end->y - start->y);
			y = max->y;
		} else if (C & C_BOTTOM) {
			x = start->x + (end->x - start->x) *
			    (0 - start->y) / (end->y - start->y);
			y = 0;
		} else if (C & C_RIGHT) {
			x = max->x;
			y = start->y + (end->y - start->y) *
			    (max->x - start->x) / (end->x - start->x);
		} else {
			x = 0;
			y = start->y + (end->y - start->y) *
			    (0 - start->x) / (end->x - start->x);
		}
		
		end->x = x;
		end->y = y;

		/* This algorithm is not optimal, multiple clippings
		 * may be required.
		 */
		C = ClipCode(end, max);
	} while (C);

	result.x = x;
	result.y = y;

	return (&result);
}

double
dag_repulse_center_compute(struct dag_point *center)
{
	struct dag_node *node, tmp;
	struct dag_point max, endp, *result;
	double force, distance, energy = 0;

	distance = repulse_center * max_attractdist;
	/* Repulse from the center of the nodes to stretch things out */
	TAILQ_FOREACH(node, &daglist, next) {
		if (node->size > 3 || !node->size) {
			node->energy = SQ(node->force.x) + SQ(node->force.y);
			node->energy = sqrt(node->energy);
			energy += node->energy;
			continue;
		}

		tmp.point = *center;
		tmp.p_x = node->p_x - tmp.p_x;
		tmp.p_y = node->p_y - tmp.p_y;

		/* If the distance is too small, ignore it */
		force = sqrt(SQ(tmp.p_x) + SQ(tmp.p_y));
		if (force < 0.01)
			continue;

		tmp.p_x /= force;
		tmp.p_y /= force;

		endp.x = node->p_x + tmp.p_x * (maxx + maxy);
		endp.y = node->p_y + tmp.p_y * (maxx + maxy);

		max.x = maxx;
		max.y = maxy;
		
		result = dag_cohensutherland_clip(&node->point, &endp, &max);

		assert(result != NULL);

		tmp.point = *result;
		tmp.size = node->size;
		dag_compute_attraction(node, &tmp, distance);

		node->energy = SQ(node->force.x) + SQ(node->force.y);
		node->energy = sqrt(node->energy);
		energy += node->energy;

		force = node->energy/node->size;
		if (force > maxforce)
			maxforce = force;
	}

	return (energy);
}

double
dag_compute_forces(void)
{
	struct dag_node *node, tmp;
	struct dag_point center;
	struct dag_edge *edge;
	double force, energy, count;

	radius = sqrt(SQ(maxx) + SQ(maxy));

	energy = 0;
	maxforce = 0;
	count = 0;
	center.x = 0;
	center.y = 0;
	TAILQ_FOREACH(node, &daglist, next) {
		double size;

		node->force.x = 0;
		node->force.y = 0;

		size = node->size;
		if (size > 2) {
			center.x += node->p_x * size;
			center.y += node->p_y * size;
			count += size;
		}

		dag_walk_region(dag_compute_repulse_cb,
		    node->p_x - 256, node->p_y - 256,
		    node->p_x + 256, node->p_y + 256,
		    node);

		SPLAY_FOREACH(edge, edgetree, &node->edgeroot) {
			if (node < edge->destination)
				continue;
			dag_compute_attraction(node, edge->destination,
			    max_attractdist);
		}

		/* Central attractor */
		size = node->size;
		if (size > 1) {
			tmp.p_x = maxx/2;
			tmp.p_y = maxy/2;
			if (dag_distance(node, &tmp) >= radius/10) {
				tmp.size = attract_center * size;
				dag_compute_repulse(node, &tmp,
				    min_repulsedist * 10);
			}
		}

		/* Attraction from four sides */
		tmp.p_x = node->p_x;
		tmp.p_y = 0;
		tmp.size = 20;
		dag_compute_repulse(node, &tmp, min_repulsedist);

		tmp.p_x = node->p_x;
		tmp.p_y = maxy;
		tmp.size = 20;
		dag_compute_repulse(node, &tmp, min_repulsedist);

		tmp.p_x = 0;
		tmp.p_y = node->p_y;
		tmp.size = 20;
		dag_compute_repulse(node, &tmp, min_repulsedist);


		tmp.p_x = maxx;
		tmp.p_y = node->p_y;
		tmp.size = 20;
		dag_compute_repulse(node, &tmp, min_repulsedist);

		node->energy = SQ(node->force.x) + SQ(node->force.y);
		node->energy = sqrt(node->energy);
		energy += node->energy;

		force = node->energy/node->size;
		if (force > maxforce)
			maxforce = force;

	}

	/* Computer mass center */
	center.x /= count;
	center.y /= count;

	if (repulse_center != 0)
		energy = dag_repulse_center_compute(&center);

	return (energy);
}

void
dag_move_node(struct dag_node *node, struct dag_point *point)
{
	struct dag_node tmp;
	int ox, oy, nx, ny;

	CLIP(point->x, maxx);
	CLIP(point->y, maxy);

	tmp.point = *point;
	while (SPLAY_FIND(nodetree, &noderoot, &tmp) != NULL) {
		int off;
		tmp.point = *point;
		CLIP(tmp.p_x, maxx);
		CLIP(tmp.p_y, maxy);

		off = (arc4random() % 8192) - 4096;
		tmp.p_x += off / 1400.0;
		off = (arc4random() % 8192) - 4096;
		tmp.p_y += off / 1400.0;

		CLIP(tmp.p_x, maxx);
		CLIP(tmp.p_y, maxy);
	}

	SPLAY_REMOVE(nodetree, &noderoot, node);

	/* Update hash */
	dag_node_hash(&node->point, &ox, &oy);
	dag_node_hash(&tmp.point, &nx, &ny);

	if (ox != nx || oy != ny) {
		TAILQ_REMOVE(&buckets[ox][oy], node, bucket_next);
		TAILQ_INSERT_TAIL(&buckets[nx][ny], node, bucket_next);
	}

	node->point = tmp.point;

	SPLAY_INSERT(nodetree, &noderoot, node);
}

void
dag_apply_forces(void)
{
	struct dag_point point;
	struct dag_node *node;
	double fact, size;
	
	/* fact = maxx / 40.0 / maxforce; */
	fact = stepsize/2;

	TAILQ_FOREACH(node, &daglist, next) {
		node->save_point = node->point;
		node->save_energy = node->energy;

		size = node->size;
		if (size < 1)
			size = 1;

		point.x = node->p_x + node->force.x * fact / size;
		point.y = node->p_y + node->force.y * fact / size;

		dag_move_node(node, &point);
	}
}

void
dag_find_minimum(void)
{
	struct dag_point point, force;
	struct dag_node *node;
	double fact, size;
	
	TAILQ_FOREACH(node, &daglist, next) {
		fact = stepsize;

		size = node->size;
		if (size < 1)
			size = 1;

		force.x = node->force.x + node->old_force.x/2;
		force.y = node->force.y + node->old_force.y/2;

		point.x = node->save_point.x + force.x * fact / size;
		point.y = node->save_point.y + force.y * fact / size;

		node->old_force.x = node->old_force.x/2 + node->force.x;
		node->old_force.y = node->old_force.y/2 + node->force.y;

		dag_move_node(node, &point);
	}
}

void
dag_update_max(int x, int y)
{
	int change = 0;

	if (x > maxx) {
		maxx = x;
		change = 1;
	}
	if (y > maxy) {
		maxy = y;
		change = 1;
	}
	if (change)
		need_rehash = 1;
}

void
dag_set_max(int x, int y)
{
	maxx = x;
	maxy = y;

	need_rehash = 1;
}
